import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from .permission_base import PermissionBase
from .user import User


class PermissionRequest(PermissionBase):
    @property
    def user(self):
        return User(self._permission.user)


class ManageablePermissionRequest(PermissionRequest):
    @property
    def requestor(self):
        """User who raised the specific permission request."""
        return User(self._permission.user)

    def approve(self, reason):
        """Approve a permission request.

        Args:
            reason: (str) A reason for permission request approval.

        Typical example:
            manageable_request.approve("it will be fun")

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-permission-requests-from-other-users
        """
        request = pb.ApprovePendingPermissionRequest(permission_id=self._permission.id, reason=reason)
        self._stub.ApprovePendingPermission(request)

    def reject(self, reason):
        """Reject a permission request.

        Args:
            reason: (str) A reason for permission request rejection.

        Typical example:
            manageable_request.reject("it's not ready yet")

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/permissions.html#managing-permission-requests-from-other-users
        """
        request = pb.RejectPendingPermissionRequest(permission_id=self._permission.id, reason=reason)
        self._stub.RejectPendingPermission(request)



from typing import List

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..entities.backfill_job import BackfillJob
from ..entities.base_job import BaseJob
from ..entities.compute_recommendation_classifiers_job import ComputeRecommendationClassifiersJob
from ..entities.compute_statistics_job import ComputeStatisticsJob
from ..entities.create_ml_dataset_job import CreateMLDatasetJob
from ..entities.extract_schema_job import ExtractSchemaJob
from ..entities.ingest_job import IngestJob
from ..entities.materialization_online_job import MaterializationOnlineJob
from ..entities.retrieve_job import RetrieveJob
from ..entities.revert_ingest_job import RevertIngestJob


class Jobs:
    def __init__(self, stub):
        self._stub = stub

    @staticmethod
    def _create_job(stub, job_proto):
        job_id = pb.JobId(job_id=job_proto.job_id)
        if job_proto.job_type == pb.JobType.Ingest:
            return IngestJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.ExtractSchema:
            return ExtractSchemaJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.Retrieve:
            return RetrieveJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.MaterializationOnline:
            return MaterializationOnlineJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.ComputeStatistics:
            return ComputeStatisticsJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.Revert:
            return RevertIngestJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.ComputeRecommendationClassifiers:
            return ComputeRecommendationClassifiersJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.CreateMLDataset:
            return CreateMLDatasetJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.Backfill:
            return BackfillJob(stub, job_id)

    def list(self, active=True, job_type=pb.JobType.Unknown) -> List[BaseJob]:
        """Return currently running jobs.

        By default, only active jobs are returned.

        Args:
            active: (bool) If True, allows active jobs.
            job_type: (JobType) Object represents a specific job type.
              INGEST | RETRIEVE | EXTRACT_SCHEMA | REVERT_INGEST | MATERIALIZATION_ONLINE | COMPUTE_STATISTICS |
              COMPUTE_RECOMMENDATION_CLASSIFIERS | CREATE_ML_DATASET | BACKFILL

        Returns:
            list[BaseJob]: A collection of jobs.

            For example:

            [Job(id=test123, type=ExtractSchema, done=True, childJobIds=[]),
            Job(id=test456, type=CreateMLDataset, done=True, childJobIds=[])]

        Typical example:
            client.jobs.list()
            client.jobs.list(active=False, job_type=INGEST)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/jobs_api.html#listing-jobs
        """
        request = pb.ListJobsRequest(active=active, job_type=job_type)
        resp = self._stub.ListJobs(request)
        return [Jobs._create_job(self._stub, job_proto) for job_proto in resp.jobs]

    def get(self, job_id: str) -> BaseJob:
        """Obtain an existing job.

        Args:
            job_id: (str) A unique id of an existing job.

        Returns:
            BaseJob: A job.

            For example:

            Job(id=test123, type=ExtractSchema, done=True, childJobIds=[])

        Typical example:
            job = client.jobs.get("job_id")

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/jobs_api.html#getting-a-job
        """
        request = pb.JobId(job_id=job_id)
        job_proto = self._stub.GetJob(request)
        return Jobs._create_job(self._stub, job_proto)

    def __repr__(self):
        return "This class wraps together methods working with jobs"



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from featurestore.core import interactive_console

from ..data_source_wrappers import get_source
from ..retrieve_holder import RetrieveHolder
from ..utils import Utils
from .revert_ingest_job import RevertIngestJob
from .user import User


class Ingest:
    def __init__(self, stub, feature_set, ingest):
        self._stub = stub
        self._feature_set = feature_set
        self._ingest = ingest

    def retrieve(self):
        """Retrieve data.

        Returns:
            RetrieveHolder: Returns a link as output for reference.

        Typical example:
            fs.retrieve()
        """
        return RetrieveHolder(self._stub, self._feature_set, "", "", self._ingest.ingest_id)

    @interactive_console.record_stats
    def revert(self):
        """Revert to a specific ingest.

        Reverting creates a new minor version with the data corresponding to the specific ingest removed.

        Typical example:
            ingest.revert()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/ingest_history_api.html#reverting-ingestion
        """
        job = self.revert_async()
        return job.wait_for_result()

    def revert_async(self) -> RevertIngestJob:
        """Create a revert ingestion job for feature set.

        Returns:
            RevertIngestJob: A job for reverting ingestion.

            A job is created with a unique id and type Revert. For example:

            Job(id=<job_id>, type=Revert, done=False, childJobIds=[])

        Typical example:
            ingest.revert_async()

        Raises:
            Exception: Manual revert is not allowed on derived feature set.
        """
        if self._feature_set.derived_from.HasField("transformation"):
            raise Exception("Manual revert is not allowed on derived feature set")

        request = pb.StartRevertIngestJobRequest()
        request.feature_set_id = self._feature_set.id
        request.feature_set_version = self._feature_set.version
        request.ingest_id = self._ingest.ingest_id
        job_id = self._stub.StartRevertIngestJob(request)
        return RevertIngestJob(self._stub, job_id)

    @property
    def ingested_records_count(self):
        return int(self._ingest.ingested_records_count)

    @property
    def ingested_at(self):
        return Utils.timestamp_to_string(self._ingest.ingest_timestamp)

    @property
    def source(self):
        return get_source(self._ingest.source)

    @property
    def scope(self):
        return Scope(self._ingest.scope)

    @property
    def started_by(self):
        return User(self._ingest.started_by)

    def __repr__(self):
        return Utils.pretty_print_proto(self._ingest)


class Scope:
    def __init__(self, scope):
        self._scope = scope

    @property
    def start_date_time(self):
        return Utils.timestamp_to_string(self._scope.start_date_time)

    @property
    def end_date_time(self):
        return Utils.timestamp_to_string(self._scope.end_date_time)

    def __repr__(self):
        return Utils.pretty_print_proto(self._scope)

    def __str__(self):
        return (
            f"Start date & time : {self._scope.start_date_time} \n"
            f"End date & time   : {self._scope.end_date_time} \n"
        )



from enum import Enum

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb


class AccessType(Enum):
    OWNER = 1
    EDITOR = 2
    CONSUMER = 3
    SENSITIVE_CONSUMER = 4
    VIEWER = 5

    @classmethod
    def from_proto_permission(cls, proto_permission_type):
        return {
            pb.PermissionType.Owner: cls.OWNER,
            pb.PermissionType.Editor: cls.EDITOR,
            pb.PermissionType.Consumer: cls.CONSUMER,
            pb.PermissionType.SensitiveConsumer: cls.SENSITIVE_CONSUMER,
            pb.PermissionType.Viewer: cls.VIEWER,
        }[proto_permission_type]

    @classmethod
    def to_proto_permission(cls, access_type):
        return {
            cls.OWNER: pb.PermissionType.Owner,
            cls.EDITOR: pb.PermissionType.Editor,
            cls.CONSUMER: pb.PermissionType.Consumer,
            cls.SENSITIVE_CONSUMER: pb.PermissionType.SensitiveConsumer,
        }[access_type]

    @classmethod
    def from_proto_active_permission(cls, active_permission):
        return {
            pb.ACTIVE_PERMISSION_NONE: None,
            pb.ACTIVE_PERMISSION_OWNER: cls.OWNER,
            pb.ACTIVE_PERMISSION_EDITOR: cls.EDITOR,
            pb.ACTIVE_PERMISSION_CONSUMER: cls.CONSUMER,
            pb.ACTIVE_PERMISSION_SENSITIVE_CONSUMER: cls.SENSITIVE_CONSUMER,
            pb.ACTIVE_PERMISSION_VIEWER: cls.VIEWER,
        }[active_permission]



from ..utils import Utils


class User:
    def __init__(self, user):
        self._user = user

    @property
    def id(self):
        return self._user.id

    @property
    def name(self):
        return self._user.name

    @property
    def email(self):
        return self._user.email

    def __repr__(self):
        return Utils.pretty_print_proto(self._user)

    def __str__(self):
        return f"Name          : {self.name} \n" f"Email         : {self.email} \n"



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from functools import reduce
from typing import TYPE_CHECKING, List, Optional, Union

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from . import feature_ref as fref
from . import feature_set_ref as fsref
from .transformation_function import TransformationFunction, TransformedFeature

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from .feature import Feature
    from .feature_set import FeatureSet


class SelectedFeatureWithTransformation:
    def __init__(self, feature: fref.FeatureRef, transformation: Optional[TransformationFunction]):
        self._feature: fref.FeatureRef = feature
        self._transformation = transformation

    @property
    def feature_set_ref(self):
        return self._feature.feature_set_ref

    @property
    def name(self):
        return self._feature.name

    @property
    def transformation(self):
        return self._transformation.function_name if self._transformation else None

    @classmethod
    def create_from(
        cls,
        feature: Union[Feature, SelectedFeatureWithTransformation, TransformedFeature],
    ):
        if isinstance(feature, SelectedFeatureWithTransformation):
            return feature
        if isinstance(feature, TransformedFeature):
            return SelectedFeatureWithTransformation(feature.feature, feature.transformation)

        return SelectedFeatureWithTransformation(feature._reference(), None)

    def __repr__(self):
        if self.transformation:
            return f"{self._transformation.function_name}({self._feature})"
        else:
            return f"{self._feature}"


class Query:
    def __init__(
        self,
        select: SelectedFrom,
        joins: Optional[list[Join]] = None,
        filter: Optional[Condition] = None,
    ):
        self._select = select
        if joins:
            self._joins = joins
        else:
            self._joins = []
        self._filter = filter

    def __repr__(self):
        return f"select: {self._select}, joins: {self._joins}, filter: {self._filter}"

    @classmethod
    def select(
        cls,
        features: list[Union[Feature, SelectedFeatureWithTransformation, TransformedFeature]],
    ):
        return Select([SelectedFeatureWithTransformation.create_from(feature) for feature in features])

    def filter(self, condition: Condition):
        if self._filter:
            new_condition = self._filter & condition
        else:
            new_condition = condition

        return Query(self._select, self._joins, new_condition)

    def _join(self, feature_set: Union[FeatureSet, fsref.FeatureSetRef], alias, join_type: JoinType) -> OpenedJoin:
        return OpenedJoin(
            self,
            feature_set if isinstance(feature_set, fsref.FeatureSetRef) else feature_set._reference(),
            alias,
            join_type,
        )

    def join(self, feature_set: Union[FeatureSet, fsref.FeatureSetRef], alias) -> OpenedJoin:
        return self._join(feature_set, alias, JoinType.INNER)

    def left_join(self, feature_set: Union[FeatureSet, fsref.FeatureSetRef], alias) -> OpenedJoin:
        return self._join(feature_set, alias, JoinType.LEFT)

    def _to_proto(self) -> pb.FeatureQuery:

        fs_to_alias = dict(
            [(join._feature_set, join._alias) for join in self._joins if join._alias]
            + [(self._select.feature_set, self._select.alias)]
            if self._select.alias
            else []
        )

        return pb.FeatureQuery(
            selected_features=[
                pb.SelectedFeature(
                    prefix=fs_to_alias.get(feature.feature_set_ref, ""),
                    feature=feature.name,
                    transformation=feature.transformation,
                )
                for feature in self._select.features
            ],
            from_feature_set=pb.FeatureQuery.From(
                feature_set=pb.VersionedId(
                    id=self._select.feature_set.id,
                    major_version=self._select.feature_set.major_version,
                ),
                alias=self._select.alias,
            ),
            where=self._filter._to_proto(fs_to_alias) if self._filter else None,
            join=[join._to_proto(fs_to_alias) for join in self._joins],
        )

    @classmethod
    def _from_proto(cls, feature_query: pb.FeatureQuery) -> Query:
        def to_feature_set_ref(version_id: pb.VersionedId) -> fsref.FeatureSetRef:
            return fsref.FeatureSetRef(version_id.id, version_id.major_version)

        def to_feature_ref(
            selected_feature: pb.SelectedFeature,
            aliases: dict[str, fsref.FeatureSetRef],
        ) -> fref.FeatureRef:
            return fref.FeatureRef(selected_feature.feature, aliases.get(selected_feature.prefix))

        from_feature_set = to_feature_set_ref(feature_query.from_feature_set.feature_set)

        alias_to_fs = dict(
            [(feature_query.from_feature_set.alias, from_feature_set)]
            if len(feature_query.from_feature_set.alias) > 0
            else []
            + [(join.alias, to_feature_set_ref(join.feature_set)) for join in feature_query.join if len(join.alias) > 0]
        )

        selected_features = [
            SelectedFeatureWithTransformation(
                to_feature_ref(feature, alias_to_fs),
                TransformationFunction(feature.transformation) if feature.transformation else None,
            )
            for feature in feature_query.selected_features
        ]

        def join_proto(query: Query, join: pb.FeatureQuery.Join) -> Query:
            if join.join_type == pb.FeatureQueryJoinType.FEATURE_QUERY_JOIN_INNER:
                join_type = JoinType.INNER
            elif join.join_type == pb.FeatureQueryJoinType.FEATURE_QUERY_JOIN_LEFT:
                join_type = JoinType.LEFT
            else:
                raise Exception(f"Unsupported join type {join.join_type}")

            joined = query._join(to_feature_set_ref(join.feature_set), join.alias, join_type).on(
                to_feature_ref(join.on[0].left, alias_to_fs),
                to_feature_ref(join.on[0].right, alias_to_fs),
            )
            return reduce(
                lambda j, on: j.on(
                    to_feature_ref(on.left, alias_to_fs),
                    to_feature_ref(on.right, alias_to_fs),
                ),
                join.on[1:],
                joined,
            ).end()

        query_after_join = reduce(
            join_proto,
            feature_query.join,
            Query.select(selected_features).from_feature_set(from_feature_set, feature_query.from_feature_set.alias),
        )

        if feature_query.HasField("where"):
            return query_after_join.filter(Condition._from_proto(feature_query.where, alias_to_fs))
        else:
            return query_after_join


class Select:
    def __init__(self, features: list[SelectedFeatureWithTransformation]):
        self._features = features

    def from_feature_set(
        self,
        feature_set: Union[FeatureSet, fsref.FeatureSetRef],
        alias: Optional[str] = None,
    ):
        return Query(
            select=SelectedFrom(
                self._features,
                feature_set if isinstance(feature_set, fsref.FeatureSetRef) else feature_set._reference(),
                alias,
            )
        )


class SelectedFrom:
    def __init__(
        self,
        features: list[SelectedFeatureWithTransformation],
        feature_set: fsref.FeatureSetRef,
        alias: Optional[str],
    ):
        self.features = features
        self.feature_set = feature_set
        self.alias = alias

    def __repr__(self):
        return f"features: {self.features}, feature_set: {self.feature_set}"


class LogicOperator(Enum):
    AND = pb.FeatureFilterLogicOperator.FEATURE_FILTER_LOGIC_OPERATOR_AND
    OR = pb.FeatureFilterLogicOperator.FEATURE_FILTER_LOGIC_OPERATOR_OR

    @classmethod
    def _from_proto(cls, proto_value: pb.FeatureFilterLogicOperator.ValueType) -> LogicOperator:
        found_values = [member for name, member in LogicOperator.__members__.items() if member.value == proto_value]
        if len(found_values) < 1:
            raise Exception(f"Not supported value {proto_value}")
        return found_values[0]


class FilterOperator(Enum):
    EQ = pb.FeatureFilterOperator.FEATURE_FILTER_OPERATOR_EQ
    NE = pb.FeatureFilterOperator.FEATURE_FILTER_OPERATOR_NE
    LT = pb.FeatureFilterOperator.FEATURE_FILTER_OPERATOR_LT
    LE = pb.FeatureFilterOperator.FEATURE_FILTER_OPERATOR_LE
    GT = pb.FeatureFilterOperator.FEATURE_FILTER_OPERATOR_GT
    GE = pb.FeatureFilterOperator.FEATURE_FILTER_OPERATOR_GE

    @classmethod
    def _from_proto(cls, proto_value: pb.FeatureFilterOperator.ValueType) -> FilterOperator:
        found_values = [member for name, member in FilterOperator.__members__.items() if member.value == proto_value]
        if len(found_values) < 1:
            raise Exception(f"Not supported value {proto_value}")
        return found_values[0]


class Condition(ABC):
    def __and__(self, condition: Condition) -> Condition:
        return LogicCondition(self, condition, LogicOperator.AND)

    def __or__(self, condition: Condition) -> Condition:
        return LogicCondition(self, condition, LogicOperator.OR)

    @abstractmethod
    def _to_proto(self, aliases: dict[fsref.FeatureSetRef, str]):
        raise NotImplementedError("Method `_to_proto` needs to be implemented by the child class")

    @classmethod
    def _from_proto(cls, where: pb.FilterConditionUnion, aliases: dict[str, fsref.FeatureSetRef]) -> Condition:
        if where.HasField("filter"):
            return FilterCondition(
                fref.FeatureRef(
                    where.filter.feature.feature,
                    aliases.get(where.filter.feature.prefix),
                ),
                FilterOperator._from_proto(where.filter.operator),
                where.filter.value,
            )
        else:
            return LogicCondition(
                Condition._from_proto(where.logic.left, aliases),
                Condition._from_proto(where.logic.right, aliases),
                LogicOperator._from_proto(where.logic.logic_operator),
            )


class LogicCondition(Condition):
    def __init__(
        self,
        left_condition: Condition,
        right_condition: Condition,
        operator: LogicOperator,
    ):
        self._left_condition = left_condition
        self._right_condition = right_condition
        self._operator = operator

    def __repr__(self):
        return f"{self._left_condition} {self._operator} {self._right_condition}"

    def _to_proto(self, aliases: dict[fsref.FeatureSetRef, str]):
        pb.FilterConditionUnion(
            logic=pb.FeatureFilterLogic(
                left=self._left_condition._to_proto(aliases),
                right=self._right_condition._to_proto(aliases),
                logic_operator=self._operator.value,
            )
        )


class FilterCondition(Condition):
    def __init__(self, selected_feature: fref.FeatureRef, operator: FilterOperator, value: str):
        self._selected_feature = selected_feature
        self._operator = operator
        self._value = value

    def __repr__(self):
        return f"{self._selected_feature} {self._operator} {self._value}"

    def _to_proto(self, aliases: dict[fsref.FeatureSetRef, str]) -> pb.FilterConditionUnion:
        return pb.FilterConditionUnion(
            filter=pb.FeatureFilter(
                feature=pb.SelectedFeature(
                    prefix=aliases.get(self._selected_feature.feature_set_ref, ""),
                    feature=(self._selected_feature.name),
                ),
                operator=self._operator.value,
                value=self._value,
            )
        )


class On:
    def __init__(
        self,
        left_feature: Union[Feature, fref.FeatureRef],
        right_feature: Union[Feature, fref.FeatureRef],
    ):
        if isinstance(left_feature, fref.FeatureRef):
            self._left_feature = left_feature
        else:
            self._left_feature = left_feature._reference()
        if isinstance(right_feature, fref.FeatureRef):
            self._right_feature = right_feature
        else:
            self._right_feature = right_feature._reference()

    def __repr__(self):
        return f"{self._left_feature} == {self._right_feature}"


class OpenedJoin:
    def __init__(self, query, feature_set: fsref.FeatureSetRef, alias: str, join_type: JoinType):
        self.__query = query
        self.__feature_set = feature_set
        self.__alias = str(alias or None)
        self.__join_type = join_type

    def on(
        self,
        left_feature: Union[Feature, fref.FeatureRef],
        right_feature: Union[Feature, fref.FeatureRef],
    ) -> Join:
        return Join(self.__query, self.__feature_set, self.__alias, [On(left_feature, right_feature)], self.__join_type)


class JoinType(Enum):
    INNER = 0
    LEFT = 1

    def _to_proto(self):
        if self == JoinType.INNER:
            return pb.FeatureQueryJoinType.FEATURE_QUERY_JOIN_INNER
        elif self == JoinType.LEFT:
            return pb.FeatureQueryJoinType.FEATURE_QUERY_JOIN_LEFT
        else:
            raise Exception(f"Unsupported type {self.name}")


class Join:
    def __init__(
        self, query: Query, feature_set: fsref.FeatureSetRef, alias: Optional[str], on: List[On], join_type: JoinType
    ):
        self._query = query
        self._feature_set = feature_set
        self._alias = str(alias or None)
        self._on = on
        self._join_type = join_type

    def __repr__(self):
        return f"feature_set: {self._feature_set}, alias: {self._alias}, on: {self._on}, join_type: {self._join_type}"

    def on(
        self,
        left_feature: Union[Feature, fref.FeatureRef],
        right_feature: Union[Feature, fref.FeatureRef],
    ) -> Join:
        return Join(
            self._query, self._feature_set, self._alias, self._on + [(On(left_feature, right_feature))], self._join_type
        )

    def end(self) -> Query:
        return Query(self._query._select, self._query._joins + [self], self._query._filter)

    def _to_proto(self, aliases: dict[fsref.FeatureSetRef, str]) -> pb.FeatureQuery.Join:
        return pb.FeatureQuery.Join(
            feature_set=pb.VersionedId(id=self._feature_set.id, major_version=self._feature_set.major_version),
            alias=self._alias,
            on=[
                pb.FeatureQuery.Join.On(
                    left=pb.SelectedFeature(
                        prefix=aliases.get(on._left_feature.feature_set_ref, ""),
                        feature=on._left_feature.name,
                    ),
                    right=pb.SelectedFeature(
                        prefix=aliases.get(on._right_feature.feature_set_ref, ""),
                        feature=on._right_feature.name,
                    ),
                )
                for on in self._on
            ],
            join_type=self._join_type._to_proto(),
        )



from typing import Optional

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from ..entities.feature_view import FeatureView


class FeatureViews:
    def __init__(self, stub: CoreServiceStub, project):
        self._project = project
        self._stub = stub

    def create(self, name: str, query, description: str = ""):
        """Create a feature view.

        Args:
            name: (str) Title of new feature view.
            query: (Query) Object represents Query instance from featurestore.core.entities.query.
            description: (str) Description about new feature view.

        Returns:
            FeatureView: A feature view with relevant metadata.

        Typical example:
            feature_view = project.feature_views.create(name = "test", description="", query)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_view_api.html#creating-a-feature-view
        """
        request = pb.CreateFeatureViewRequest(
            name=name,
            description=description,
            project_id=self._project.id,
            query=query._to_proto(),
        )
        response = self._stub.CreateFeatureView(request)
        return FeatureView(self._stub, response.feature_view)

    def get(self, name: str, version: Optional[int] = None):
        """Obtain an existing feature view.

        Args:
            name: (str) Title of existing feature view.
            version: (int) Version of existing feature view.

        Returns:
            FeatureView: A feature view with relevant metadata.

        Typical example:
            feature_view = project.feature_views.get("feature_view_name", version=None)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_view_api.html#obtaining-a-feature-view
        """
        request = pb.GetFeatureViewRequest(project_id=self._project.id, name=name, version=(version or 0))
        resource = self._stub.GetFeatureView(request)
        return FeatureView(self._stub, resource.feature_view)

    def list(self):
        """Return a collection of feature views within a project.

        Returns:
            list[FeatureView]: A list of existing feature views.

        Typical example:
            project.feature_views.list()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_view_api.html#listing-feature-views-within-a-project
        """
        request = pb.ListFeatureViewsRequest(project_id=self._project.id)
        response = self._stub.ListFeatureViews(request)
        return [FeatureView(self._stub, feature_view) for feature_view in response.feature_views]



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..utils import Utils
from .project import Project


class RecentlyUsedProject:
    def __init__(self, stub, recently_used_project):
        self._stub = stub
        self._recently_used_project = recently_used_project

    @property
    def name(self):
        return self._recently_used_project.project_name

    @property
    def description(self):
        return self._recently_used_project.project_description

    @property
    def updated_at(self):
        return Utils.timestamp_to_string(self._recently_used_project.updated_at)

    @property
    def last_access_at(self):
        return Utils.timestamp_to_string(self._recently_used_project.last_access_at)

    def get_project(self):
        request = pb.GetProjectRequest(project_name=self._recently_used_project.project_name)
        response = self._stub.GetProject(request)
        return Project(self._stub, response.project)

    def __repr__(self):
        return Utils.pretty_print_proto(self._recently_used_project)



import base64
import datetime
import hashlib
import json
import os
import tempfile
import textwrap
import warnings
from typing import Optional

import requests
from dateutil import tz
from google.protobuf import json_format


class Utils:
    @staticmethod
    def pretty_print_proto(m):
        return json_format.MessageToJson(m, including_default_value_fields=True)

    @staticmethod
    def proto_to_dict(m):
        return json_format.MessageToDict(m, including_default_value_fields=True, preserving_proto_field_name=True)

    @staticmethod
    def output_indent_spacing(m, space):
        return textwrap.indent(m, space)

    @staticmethod
    def timestamp_to_string(timestamp):
        if timestamp.ByteSize() != 0:
            return timestamp.ToDatetime().isoformat()
        else:
            return None

    @staticmethod
    def is_running_on_databricks() -> bool:
        try:
            from pyspark.sql import SparkSession

            spark = SparkSession.getActiveSession()
            spark.conf.get("spark.databricks.clusterUsageTags.sparkVersion")
            return True
        except Exception:
            return False

    @staticmethod
    def read_env(variable_name, source):
        value = os.environ.get(variable_name)
        if value is None:
            raise Exception(
                "Environment variable "
                + variable_name
                + " is missing, it is required to read from "
                + source
                + " data source."
            )
        else:
            return value

    @staticmethod
    def download_files(output_dir, download_links):
        if output_dir is None:
            output_dir = tempfile.mkdtemp()
        for idx, path in list(enumerate(download_links)):
            dest = os.path.join(output_dir, "part_" + str(idx) + ".parquet")
            Utils.download_file(dest, path)
        return output_dir

    @staticmethod
    def download_file(dest, origin):
        with requests.get(origin, stream=True) as r:
            r.raise_for_status()
            with open(dest, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return dest

    @staticmethod
    def fetch_preview_as_json_array(preview_url: str):
        with requests.get(preview_url) as r:
            r.raise_for_status()
            preview_array = r.text.strip().split("\n")
            return [json.loads(item) for item in preview_array]

    @staticmethod
    def generate_md5_checksum(file):
        hash_md5 = hashlib.md5()
        with open(file, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return base64.b64encode(hash_md5.digest())

    @staticmethod
    def date_time_to_proto_timestamp(date_time: Optional[datetime.datetime]):
        if date_time is None:
            return None
        else:
            from google.protobuf.timestamp_pb2 import Timestamp

            t = date_time.timestamp()
            seconds = int(t)
            nanos = int(t % 1 * 1e9)

            return Timestamp(seconds=seconds, nanos=nanos)

    @staticmethod
    def convert_timestamp_to_str_with_zone(timestamp):
        from_zone = tz.gettz("UTC")
        client_zone = tz.tzlocal()
        return timestamp.ToDatetime().replace(tzinfo=from_zone).astimezone(client_zone).strftime("%Y-%m-%dT%H:%M:%S %z")

    @staticmethod
    def warn_deprecated(message):
        warnings.warn(message, DeprecationWarning, stacklevel=2)
